package ru.amse.kiselev.fsmeditor.graphics;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Rectangle2D;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.JComponent;

import ru.amse.kiselev.fsmeditor.automaton.IAlphabet;
import ru.amse.kiselev.fsmeditor.automaton.IState;
import ru.amse.kiselev.fsmeditor.automaton.ITransition;
import ru.amse.kiselev.fsmeditor.graphics.representation.IGraphicalRepresentation;
import ru.amse.kiselev.fsmeditor.graphics.representation.IVertex;
import ru.amse.kiselev.fsmeditor.interpreter.IInterpreter;

/**
 * @author Yury Kiselev
 * 
 * @param <A>
 */
public class AutomatonComponent<A extends IAlphabet> extends JComponent {
	private static final long serialVersionUID = 1L;

	public static final Font myLabelFont = new Font("Arial", Font.PLAIN, 18);
	private IGraphicalRepresentation<A> myGraphicalRepresentation;
	private IInterpreter<A> myInterpreter;
	private IVertex<A> myMovedOnVertex;
	private A myAlphabet;
	private Map<ITransition<A>, Shape> myTrLabelBounds;

	private boolean myStraightNeeded;
	private IVertex<A> myStart;
	private Point myEnd;

	public AutomatonComponent(IGraphicalRepresentation<A> gr,
			IInterpreter<A> interpreter, A alphabet) {
		myGraphicalRepresentation = gr;
		myInterpreter = interpreter;
		myAlphabet = alphabet;

		myTrLabelBounds = new HashMap<ITransition<A>, Shape>();
	}

	public void setData(IInterpreter<A> interpreter,
			IGraphicalRepresentation<A> graphicalRepresentation) {
		myStraightNeeded = false;
		myStart = null;
		myEnd = null;

		myMovedOnVertex = null;
		myTrLabelBounds.clear();

		myInterpreter = interpreter;
		myGraphicalRepresentation = graphicalRepresentation;
	}

	public IInterpreter<A> getInterpreter() {
		return myInterpreter;
	}

	/**
	 * transitions working adding to the Map not needed -- expected during
	 * redrawing
	 * 
	 * @param transition
	 */
	public void addTransition(ITransition<A> transition) {
		myInterpreter.getAutomaton().addTransition(transition);
	}

	public void removeTransition(ITransition<A> transition) {
		myTrLabelBounds.remove(transition);
		myInterpreter.getAutomaton().removeTransition(transition);
	}

	public void setHighlightedVertex(IVertex<A> vertex) {
		myMovedOnVertex = vertex;
	}

	public void drawTransition(IVertex<A> start, IVertex<A> end, Graphics gc) {
		Graphics2D gc2 = (Graphics2D) gc;
		IState<A> sEnd = end.getState();
		IState<A> sStart = start.getState();
		if (sStart.connectedWith(sEnd).size() == 0) {
			return;
		}

		Collection<ITransition<A>> inversedTransitionCollection = sEnd
				.connectedWith(sStart);

		if (inversedTransitionCollection.size() == 0) {
			AffineTransform at = Drawing.drawStraightPointer(gc, start.getX(),
					start.getY(), end.getX(), end.getY(), Color.black, start
							.getRadius(), end.getRadius());
			FontMetrics metrics = gc2.getFontMetrics();
			int height = metrics.getHeight();
			drawText(gc, at, Color.black, sStart, sEnd, -height / 2 + 2);
		} else if (start == end) {
			AffineTransform at = drawArcLine(gc, start);
			drawText(gc, at, Color.black, sStart, sStart, - 3 * Drawing.POINT_RADIUS);
		} else {
			AffineTransform at = Drawing.drawQuadCurvePointer(gc, start.getX(),
					start.getY(), end.getX(), end.getY(), Drawing.CURVINESS,
					Color.black, start.getRadius(), end.getRadius(), 1);

			drawText(gc, at, Color.black, sStart, sEnd, 0);
		}
	}

	private void drawText(Graphics gc, AffineTransform at, Color black,
			IState<A> start, IState<A> end, int dy) {
		Graphics2D gc2 = (Graphics2D) gc;
		Collection<ITransition<A>> transitionCollection = start
				.connectedWith(end);
		int stringLength = 0;
		FontMetrics metrics = gc2.getFontMetrics();
		for (ITransition<A> tr : transitionCollection) {
			String cond = myAlphabet.getString(tr.getCondition().getSymbol());
			stringLength += metrics.stringWidth(cond) + 2
					* Drawing.CONDITION_INTERVAL;
		}
		// stringLength -= 2 * Drawing.CONDITION_INTERVAL;

		int currentLength = Drawing.CONDITION_INTERVAL;
		FontRenderContext frc = gc2.getFontRenderContext();
		int height = metrics.getHeight();
		for (ITransition<A> tr : transitionCollection) {
			String cond = myAlphabet.getString(tr.getCondition().getSymbol());
			AffineTransform past = gc2.getTransform();
			AffineTransform newTr = (AffineTransform) past.clone();
			newTr.concatenate(at);
			gc2.setTransform(newTr);

			TextLayout layout = new TextLayout(cond, myLabelFont, frc);
			int width = metrics.stringWidth(cond);
			layout.draw(gc2, -stringLength / 2 + currentLength, -height / 2
					+ metrics.getAscent() + dy);
			Rectangle2D bounds = layout.getBounds();
			bounds.setRect(bounds.getX() - stringLength / 2 + currentLength
					- Drawing.CONDITION_INTERVAL, bounds.getY() - height / 2
					+ metrics.getAscent() + dy - Drawing.BORDER_TOLERANCE,
					bounds.getWidth() + 2 * Drawing.CONDITION_INTERVAL, bounds
							.getHeight()
							+ 2 * Drawing.BORDER_TOLERANCE);
			//gc2.draw(bounds);

			myTrLabelBounds.put(tr, at.createTransformedShape(bounds));

			gc2.setTransform(past);
			currentLength += width + 2 * Drawing.CONDITION_INTERVAL;
		}

	}

	public void drawTransitions(Graphics gc) {
		Collection<IVertex<A>> vertexes = myGraphicalRepresentation
				.getVertexes();
		for (Iterator<IVertex<A>> i1 = vertexes.iterator(); i1.hasNext();) {
			IVertex<A> v1 = i1.next();
			for (Iterator<IVertex<A>> i2 = vertexes.iterator(); i2.hasNext();) {
				if (i2 != i1) {
					drawTransition(v1, (IVertex<A>) i2.next(), gc);
				}				
			}
			drawTransition(v1, v1, gc);
		}
	}

	public AffineTransform drawArcLine(Graphics gc, IVertex<A> center) {
		int r = Drawing.TRANSITION_RADIUS;// + Drawing.POINT_RADIUS;
		int R = center.getRadius() + Drawing.POINT_RADIUS;
		
		
		double angleF = 2 * Math.asin(R / 2.0 / r);
		Graphics2D gc2 = (Graphics2D) gc;
		double startAngle = (angleF / Math.PI * 180 - 90) / 180 * Math.PI;
		double endAngle = (-2 * angleF / Math.PI * 180.0 + 360.0) / 180 * Math.PI;
		gc2.drawArc(center.getX() - r, 
				center.getY() - 2 * r, 
				2 * r, 
				2 * r, 
				(int) (startAngle * 180 / Math.PI), 
				(int) (endAngle * 180 / Math.PI));
		AffineTransform at = new AffineTransform();
		at.translate(center.getX(), center.getY() - 2 * r);
		
		Drawing.drawPointer(gc2, -angleF - 0.5 / Math.PI, center.getX() - Math.sin(angleF) * r,
				center.getY() - (1 - Math.cos(angleF)) * r);
		
		return at;
	}

	public void setStraightLine(boolean needed, IVertex<A> start, Point point) {
		myStraightNeeded = needed;
		myStart = start;
		myEnd = point;
	}

	public void drawStraightLine(Graphics gc, Color color) {
		Drawing.drawStraightPointer(gc, new Point(myStart.getX(), myStart
				.getY()), myEnd, color, 0, 0);
	}

	public void drawVertex(IVertex<A> v, Graphics gc) {
		Graphics2D gc2 = (Graphics2D) gc;

		if (v.getState().isFinal()) {
			gc2.setColor(Color.red);
		} else {
			gc2.setColor(Color.green);
		}
		if (v.getState() == myInterpreter.getCurrentState()) {
			gc2.setColor(Color.blue);
		}

		gc2.fill(new Ellipse2D.Double(v.getX() - v.getRadius(), v.getY()
				- v.getRadius(), v.getRadius() * 2, v.getRadius() * 2));

		if (v.getState() == myGraphicalRepresentation.getAutomaton()
				.getInitialState()) {
			gc2.setColor(Color.blue);
			gc2.fill(new Ellipse2D.Double(v.getX() + v.getRadius() / 2
					- Drawing.POINT_RADIUS, v.getY() + v.getRadius() / 2
					- Drawing.POINT_RADIUS, Drawing.POINT_RADIUS * 2,
					Drawing.POINT_RADIUS * 2));
		}

		gc2.setColor(Color.black);
		if (myMovedOnVertex == v) {
			gc2.draw(new Ellipse2D.Double(v.getX() - v.getRadius(), v.getY()
					- v.getRadius(), v.getRadius() * 2, v.getRadius() * 2));
		}

		FontMetrics metrics = gc2.getFontMetrics();
		String VLabel = v.getState().getLabel();
		int width = metrics.stringWidth(VLabel);
		int height = metrics.getHeight();
		gc2.drawString(VLabel, (int) v.getX() - width / 2, (int) v.getY()
				- height / 2 + metrics.getAscent());
	}

	public void drawAllVertexes(Graphics gc) {
		for (IVertex<A> v : myGraphicalRepresentation.getVertexes()) {
			drawVertex(v, gc);
		}
	}

	public void paint(Graphics gc) {
		((Graphics2D) gc).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);
		gc.setFont(myLabelFont);
		drawTransitions(gc);
		drawAllVertexes(gc);

		if (myStraightNeeded) {
			drawStraightLine(gc, Color.RED);
		}
		gc.setFont(Drawing.NAME_FONT);
		gc.drawString("Automaton", 15, 15);
	}

	public Collection<Entry<ITransition<A>, Shape>> getTransitionLabelsPositions() {
		return Collections.unmodifiableCollection(myTrLabelBounds.entrySet());
	}

	public void setStraightLine(boolean b) {
		myStraightNeeded = b;
	}

	public IGraphicalRepresentation<A> getGraphicalRepresentation() {
		return myGraphicalRepresentation;
	}
}
