package ru.amse.kiselev.fsmeditor.graphics;

import java.awt.Shape;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.Collection;
import java.util.Map.Entry;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import ru.amse.kiselev.fsmeditor.automaton.AutomatonFactory;
import ru.amse.kiselev.fsmeditor.automaton.IAlphabet;
import ru.amse.kiselev.fsmeditor.automaton.IState;
import ru.amse.kiselev.fsmeditor.automaton.ITransition;
import ru.amse.kiselev.fsmeditor.graphics.representation.IVertex;
import ru.amse.kiselev.fsmeditor.graphics.representation.Vertex;

public class MouseHandler<A extends IAlphabet> extends MouseAdapter {
	private MouseAdapter myMouseHandler;
	private MouseMotionsHandler myMouseMotionsHandler;
	private MouseAddingVertex myMouseAddingVertex;
	private MouseAddingTransition myMouseAddingTransition;
	private MouseSetInitial myMouseSetInitial;
	private MouseSetFinal myMouseSetFinal;
	
	private JPanel myImagePanel;
	private JFrame myFrame;
	private AutomatonFactory<A> myAutomatonFactory;
	private AutomatonComponent<A> myAutomatonComponent;
	

	public MouseHandler(JFrame frame, JPanel imagePanel,
			AutomatonFactory<A> automatonFactory,
			AutomatonComponent<A> automatonComponent) {
		myMouseMotionsHandler = new MouseMotionsHandler();
		myMouseAddingVertex = new MouseAddingVertex();
		myMouseAddingTransition = new MouseAddingTransition();
		myMouseSetInitial = new MouseSetInitial();
		myMouseSetFinal = new MouseSetFinal();

		myMouseHandler = myMouseMotionsHandler;
		myImagePanel = imagePanel;
		myFrame = frame;
		myAutomatonFactory = automatonFactory;
		myAutomatonComponent = automatonComponent;
	}

	private class MouseMotionsHandler extends MouseAdapter {
		private IVertex<A> myDraggedVertex;

		private MouseMotionsHandler() {
			myDraggedVertex = null;
		}

		public void mouseClicked(MouseEvent e) {
			IVertex<A> vertex = getVertex(e);
			if (vertex != null) {
				editVertex(vertex);
			} else {
				ITransition<A> transition = getTransition(e);
				if (transition != null) {
					editTransition(transition);
				}
			}
		}

		public void mouseDragged(MouseEvent e) {
			if (myDraggedVertex != null) {
				if (myImagePanel.contains(e.getPoint())) {
					myDraggedVertex.setLocation(e.getPoint());
				}
				myFrame.repaint();
			}
		}

		public void mousePressed(MouseEvent e) {
			IVertex<A> vertex = getVertex(e);

			if (vertex != null) {
				myDraggedVertex = vertex;
				myFrame.repaint();
			}
		}

		public void mouseReleased(MouseEvent e) {
			myDraggedVertex = null;
		}
	}// MMH (1)

	private class MouseAddingVertex extends MouseAdapter {
		public void mouseReleased(MouseEvent e) {
			IState<A> state = myAutomatonFactory.createState("New");
			IVertex<A> newVertex = new Vertex<A>((Point2D) e.getPoint(),
					IVertex.DEFAULT_RADIUS, state);

			myAutomatonComponent.getGraphicalRepresentation().addVertex(
					newVertex);
			myFrame.repaint();
			editVertex(newVertex);
			setMouseMotionsMode();
			myFrame.repaint();
		}
	}// MAV (2)

	private class MouseAddingTransition extends MouseAdapter {
		private IVertex<A> myFirstTrVertex;

		private MouseAddingTransition() {
			myFirstTrVertex = null;
		}

		public void mouseClicked(MouseEvent e) {
			myFirstTrVertex = null;
			setMouseMotionsMode();
		}

		public void mousePressed(MouseEvent e) {
			IVertex<A> vertex = getVertex(e);
			if (vertex != null) {
				myFirstTrVertex = vertex;
			} else {
				setMouseMotionsMode();
			}
		}

		public void mouseDragged(MouseEvent e) {
			myAutomatonComponent.setStraightLine(true, myFirstTrVertex,
					e.getPoint());
			myFrame.repaint();
		}

		public void mouseReleased(MouseEvent e) {
			IVertex<A> vertex = getVertex(e);
			if (vertex != null) {
				IState<A> start = myFirstTrVertex.getState();
				ITransition<A> transition = myAutomatonFactory
						.createTransition(start, vertex.getState(),
								myAutomatonFactory.createCondition(0));
				start.addTransition(transition);
				myAutomatonComponent.setStraightLine(false);
				myFrame.repaint();
				editTransition(transition);
				myFrame.repaint();
				myFirstTrVertex = null;
			} else {
				myFirstTrVertex = null;
				myAutomatonComponent.setStraightLine(false);
				myFrame.repaint();
			}
			setMouseMotionsMode();
		}
	}// MAT (3)

	private class MouseSetInitial extends MouseAdapter {
		public void mouseClicked(MouseEvent e) {
			IVertex<A> vertex = getVertex(e);
			if (vertex == null) {
				setMouseMotionsMode();
			} else {
				myAutomatonComponent.getGraphicalRepresentation()
						.getAutomaton().setInitialState(vertex.getState());
				setMouseMotionsMode();
				myImagePanel.repaint();
			}
		}
		public void mouseDragged(MouseEvent e) {
			setMouseMotionsMode();
		}
	}// MSI (4)
	
	public class MouseSetFinal extends MouseAdapter {
		public void mouseClicked(MouseEvent e) {
			IVertex<A> vertex = getVertex(e);
			if (vertex == null) {
				setMouseMotionsMode();
			} else {
				boolean isFinalState = 	vertex.getState().isFinal();
				if (isFinalState) {
					vertex.getState().setNonFinal();
				} else {
					vertex.getState().setFinal();
				}
				//setMouseMotionsMode();
				myImagePanel.repaint();
			}
		}
		public void mouseDragged(MouseEvent e) {
			setMouseMotionsMode();
		}
	}//MSF (5)

	private IVertex<A> getVertex(MouseEvent e) {
		Collection<IVertex<A>> allVertexes = myAutomatonComponent
				.getGraphicalRepresentation().getVertexes();
		for (IVertex<A> vertex : allVertexes) {
			if (vertex.distance(e.getPoint()) < vertex.getRadius()) {
				return vertex;
			}
		}
		return null;
	}

	private ITransition<A> getTransition(MouseEvent e) {
		Collection<Entry<ITransition<A>, Shape>> myTrLabelEntries = myAutomatonComponent
				.getTransitionLabelsPositions();
		for (Entry<ITransition<A>, Shape> entry : myTrLabelEntries) {
			if ((entry.getValue() != null)
					&& (entry.getValue().contains(e.getPoint()))) {
				return entry.getKey();
			}
		}
		return null;
	}
	
	private final void editVertex(IVertex<A> vertex) {
		String s = (String) JOptionPane.showInputDialog(myFrame,
				"Enter new Vertex label: \n(or empty string to delete vertex)",
				"Vertex label edition", JOptionPane.PLAIN_MESSAGE, null, null,
				vertex.getString());
		if (s == null) {
			return;
		}
		if (s.length() == 0) {
			myAutomatonComponent.getGraphicalRepresentation().removeVertex(
					vertex);
			return;
		}
		vertex.getState().setLabel(s);
	}

	private final void editTransition(ITransition<A> transition) {
		String s = (String) JOptionPane
				.showInputDialog(
						myFrame,
						"Enter new Transition label: \n(or empty string to delete transition)",
						"Vertex label edition", JOptionPane.PLAIN_MESSAGE,
						null, null, myAutomatonFactory.getAlphabet().getString(
								transition.getCondition().getSymbol()));
		if (s == null){
			myFrame.repaint();
			return;
		}
		if (s.length() == 0) {
			myAutomatonComponent.removeTransition(transition);
			myFrame.repaint();
			return;
		}
		int n = myAutomatonFactory.getAlphabet().getNum(s);
		if (n >= 0) {
			myAutomatonComponent.removeTransition(transition);
			transition = myAutomatonFactory.createTransition(transition
					.getStart(), transition.getEnd(), myAutomatonFactory
					.createCondition(n));
			transition.getStart().addTransition(transition);
		} else {
			JOptionPane.showMessageDialog(myFrame, "Sorry, \"" + s + "\" "
					+ "isn't a valid Condition.\n"
					+ "Please enter correct Condition.", "Try again",
					JOptionPane.ERROR_MESSAGE);
			editTransition(transition);
		}
		myFrame.repaint();
	}

	public void mouseDragged(MouseEvent e) {
		myMouseHandler.mouseDragged(e);
	}

	public void mouseMoved(MouseEvent e) {
		myAutomatonComponent.setHighlightedVertex(getVertex(e));
		myFrame.repaint();
	}

	public void mouseClicked(MouseEvent e) {
		myMouseHandler.mouseClicked(e);
	}

	public void mousePressed(MouseEvent e) {
		myMouseHandler.mousePressed(e);
	}

	public void mouseReleased(MouseEvent e) {
		myMouseHandler.mouseReleased(e);
	}

	public void setAddingTransitionMode() {
		myMouseHandler = myMouseAddingTransition;
	}

	public void setAddingVertexMode() {
		myMouseHandler = myMouseAddingVertex;
	}

	public void setMouseMotionsMode() {
		myMouseHandler = myMouseMotionsHandler;
	}
	public void setInitialSetMode() {
		myMouseHandler = myMouseSetInitial;
	}

	public void setFinalSetMode() {
		myMouseHandler = myMouseSetFinal;
	} 
}
